/* ***********************************************************************
 * File: fft_mpi_r2
 * Description: Parallel implementation of FFT using the D&C methology.
 * Author: Coro
 * Date: 13.05.98
 * History:
 *            modified by jlroda for BSP euroPVM/MPI 99
 * 22.10.1999 modified by jaglez. Unnecessary input data copies removed.
 *            Libraries ll_mpi2.h and my_time.h redefined.
 * 15.11.1999 Output data copies removed.
 * 18.11.1999 Using a several blocks of dinamic memory.
 * ***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ll_mpi2.h"
#include "mytime.h"
#include "complex.h"

#define K	  (1024)

void seqDandCFFT(Complex *A, Complex *a, Complex *W, int N, int stride, \
								 Complex *D);
void parDandCFFT(Complex *A, Complex *a, Complex *W, int N, int stride, \
								 Complex *D);
void Roots(int Size, Complex *W);

/* ***********************************************************************
 * Description: Computes Sequential Fast Fourier Transform
 * Parameters:
 *   Complex *A, transformed output signal
 *   Complex *a, input signal
 *   Complex *W, vector containing the roots
 *   int N, number of elements in a
 *   int stride, between consecutive elements in A to be processed
 *   Complex *D, auxiliar vector to make combination
 * ***********************************************************************/
void seqDandCFFT(Complex *A, Complex *a, Complex *W, int N, int stride,  \
								 Complex *D) {
  Complex *B, *C;
  Complex Aux, *pW;
  int i, n;

  if(N == 1) {
		A[0].re = a[0].re;
		A[0].im = a[0].im;
  }
  else {
    /* Division phase. No copy of input data is performed */
    n = (N >> 1);
    B = A;
		C = A + n;

    /* Subproblems resolution phase */
    seqDandCFFT(B, a, W, n, stride<<1, D);
    seqDandCFFT(C, a+stride, W, n, stride<<1, D);

    /* Combination phase in the auxiliar vector D */
    for(i = 0, pW = W; i < n; i++, pW += stride) {
      Aux.re = pW->re * C[i].re - pW->im * C[i].im;
      Aux.im = pW->re * C[i].im + pW->im * C[i].re;
      
      D[i].re = B[i].re + Aux.re;
      D[i].im = B[i].im + Aux.im;
      D[i+n].re = B[i].re - Aux.re;
      D[i+n].im = B[i].im - Aux.im;                                     
    }
		/* Moving the transformed signal to A */
		for(i = 0; i < N; i++) A[i] = D[i];
  }
}

/* ***********************************************************************
 * Description: Computes Parallel Fast Fourier Transform
 * Parameters:
 *   Complex *A, transformed output signal
 *   Complex *a, input signal
 *   Complex *W, vector containing the roots
 *   int N, number of elements in a
 *   int stride, between consecutive elements in A to be processed
 *   Complex *D, auxiliar vector to make combination
 * ***********************************************************************/
void parDandCFFT(Complex *A, Complex *a, Complex *W, int N, int stride,  \
								 Complex *D) {
  Complex *B, *C;
  Complex Aux, *pW;
  int i, n, size;

  if(NUMPROCESSORS > 1) {
    if(N == 1) {
      A[0].re = a[0].re;
      A[0].im = a[0].im;
    }
    else {
      /* Division phase. No copy of input data is performed */
      n = (N >> 1);
			size = n*sizeof(Complex);
      B = A;
		  C = A + n;
    
      /* Subproblems resolution phase */
      PAR(parDandCFFT(B, a, W, n, stride<<1, D), B, size,
          parDandCFFT(C, a+stride, W, n,stride<<1, D), C, size);
          
      /* Combination phase in the auxiliar vector D */
      for(i = 0, pW = W; i < n; i++, pW += stride) {
        Aux.re = pW->re * C[i].re - pW->im * C[i].im;
        Aux.im = pW->re * C[i].im + pW->im * C[i].re;
      
        D[i].re = B[i].re + Aux.re;
        D[i].im = B[i].im + Aux.im;
        D[i+n].re = B[i].re - Aux.re;
        D[i+n].im = B[i].im - Aux.im;                                     
      }
		  /* Moving the transformed signal to A */
		  for(i = 0; i < N; i++) A[i] = D[i];
    }
  }  
  else
    seqDandCFFT(A, a, W, N, stride, D);
}

/* ***********************************************************************

 * ***********************************************************************/
void Roots(int Size, Complex *W) {
  int i;
  double phi;
  Complex Omega;
  
  phi = 4 * atan(1.0) / (double)Size;                         /* PI/Size */
  Omega.re = cos(phi);
  Omega.im = sin(phi);
  W[0].re = 1;
  W[0].im = 0;
  for(i = 1; i < Size; i++) {
    W[i].re = W[i-1].re * Omega.re - W[i-1].im * Omega.im;
    W[i].im = W[i-1].re * Omega.im + W[i-1].im * Omega.re;
  }
}

/* ***********************************************************************

 * ***********************************************************************/
int main(int argc, char *argv[]) {
  int N;
  Complex *a, *A, *W, *D;
  int sizeInKb;
  double t_par, t_seq;
  CLOCK_TYPE chrono;

  INITIALIZE(argc, argv);                           /* LL Initialisation */

  if(argc > 1)                      /* Command line arguments processing */
    sizeInKb = atoi(argv[1]);
  else 
    sizeInKb = 64;
    
#ifdef DEBUG
  N = 64;
#else    
  N = K * sizeInKb;
#endif
  
  /* Memory allocation */
  a = (Complex*)calloc(N, sizeof(Complex));
  A = (Complex*)calloc(N, sizeof(Complex));
  D = (Complex*)calloc(N, sizeof(Complex));
  W = (Complex*)calloc(N>>1, sizeof(Complex));
  if((a==NULL) || (A==NULL) || (D==NULL) || (W==NULL)) EXIT(1);

  initialize(N, a);                        /* Generate test input signal */

#ifdef DEBUG                                     /* Display input vector */
  write_array(N, a);
#endif
  
  Roots(N>>1, W);            /* Initialise the vector of imaginary roots */
  
  CLOCK_Start(chrono);                                /* Sequential test */
  seqDandCFFT(A, a, W, N, 1, D);
  CLOCK_End(chrono, t_seq);

  MPI_Barrier(MPI_COMM_WORLD);

  CLOCK_Start(chrono);                                  /* Parallel test */
  parDandCFFT(A, a, W, N, 1, D);
  CLOCK_End(chrono, t_par);

#ifdef DEBUG                                    /* Display output vector */
  write_array(N, A);
#endif

  /* Display time */
  LL_printMaster("NAME\tSTATUS\tP\tN\tTIME_SEQ\tTIME_PAR\n");
  LL_printMaster("%d\t%d\t%d\t%d\t%12.8lf\t%12.8lf\n", NAME,             \
								 test_array(N, A), NUMPROCESSORS, N, t_seq, t_par);

  free(W);                                        /* Memory deallocation */
  free(D);
  free(A);
  free(a);
  
  MPI_Finalize();
  return 0;
}

